; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: OS independent micro configuration files library.
;
;  Target OS: Any
;
;  Dependencies: strlib.asm; arrays.asm;
;
;  Notes:
;
;_________________________________________________________________________________________
module "uConfig library"
struct TConfig
 .hConfig     dd ?          ; the text of the config database.
 .hCurrentDir dd ?
 .md5         dd ?          ; the MD5 hash of the original file.

 if options.Threads
   .mutex    TMutex        ; in case of multithreading program.
 end if
ends



proc CfgClean, .pConfig
begin
        pushad

        mov     esi, [.pConfig]

  if options.Threads
        lea     eax, [esi+TConfig.mutex]
        stdcall WaitForMutex, eax, -1
  end if

        xor     edx, edx

        mov     ecx, [esi+TConfig.hConfig]
        jecxz   @f
        stdcall StrDel, ecx
        mov     [esi+TConfig.hConfig], edx
@@:

        mov     ecx, [esi+TConfig.md5]
        jecxz   @f
        stdcall StrDel, ecx
        mov     [esi+TConfig.md5], edx
@@:

        mov     ecx, [esi+TConfig.hCurrentDir]
        jecxz   @f
        stdcall StrDel, ecx
        mov     [esi+TConfig.hCurrentDir], edx
@@:

  if options.Threads
        lea     eax, [esi+TConfig.mutex]
        stdcall MutexDestroy, eax
  end if

        popad
        return
endp



proc CfgLoad, .pConfig, .hFilename
begin
        pushad

        mov     edi, [.pConfig]

        stdcall LoadBinaryFile, [.hFilename]
        jc      .finish

        push    eax
        stdcall StrDup, eax
        mov     [edi+TConfig.hConfig], eax
        stdcall FreeMem ; from the stack.

        stdcall StrMD5, [edi+TConfig.hConfig]
        mov     [edi+TConfig.md5], eax

        stdcall StrNew
        mov     [edi+TConfig.hCurrentDir], eax

.finish:
  if options.Threads
        lea     eax, [edi+TConfig.mutex]
        stdcall MutexCreate, 0, eax
        stdcall MutexRelease, eax
  end if

        popad
        return
endp



proc CfgSave, .pConfig, .hFilename
begin
        pushad

        mov     esi, [.pConfig]

  if options.Threads
        lea     eax, [esi+TConfig.mutex]
        stdcall WaitForMutex, eax, -1
  end if

; it is not as it should be. The save procedure have to check the MD5 sum
; of the existing file and then to make some sort of merge, if the file
; was changed in between. That is why the field TConfig.md5 exists.

        stdcall StrLen, [esi+TConfig.hConfig]
        mov     ecx, eax
        stdcall StrPtr, [esi+TConfig.hConfig]
        stdcall SaveBinaryFile, [.hFilename], eax, ecx

.finish:
  if options.Threads
        lea     eax, [esi+TConfig.mutex]
        stdcall MutexRelease, eax
  end if
        popad
        return
endp


proc CfgChDir, .pConfig, .dir, .fRelative
begin
        pushad

        mov     esi, [.pConfig]

        cmp     [.dir], 0
        jne     .set_it

        stdcall StrDel, [esi+TConfig.hCurrentDir]
        stdcall StrNew
        mov     [esi+TConfig.hCurrentDir], eax
        popad
        return

.set_it:
        stdcall StrDup, [.dir]
        mov     edx, eax

        stdcall StrClipSpacesR, edx
        stdcall StrClipSpacesL, edx

        stdcall StrLen, edx
        mov     ecx, eax

        stdcall StrPtr, edx
        cmp     byte [eax+ecx-1], ':'
        je      @f
        stdcall StrCharCat, edx, ':'
@@:
        push    edx
        push    [esi+TConfig.hCurrentDir]

        cmp     [.fRelative], 0
        je      .abs

        stdcall StrCat
        jmp     .path_ok

.abs:
        stdcall StrCopy

.path_ok:
        stdcall StrDel, edx
        stdcall StrNormalizePath, [esi+TConfig.hCurrentDir], '::'

        popad
        return
endp




; Reads a record from a config file as an integer value.
; Returns:
;   CF=1 if the record was not found or when the record contains no number.
;     eax = -1; the record was not found.
;     eax = 0; the record exists, but contains no valid number.
;   CF=0; eax=number if the record was found and contains a number.
proc CfgGetInt, .pConfig, .path
begin
        or      eax, -1
        stdcall CfgGetRecord, [.pConfig], [.path]
        jc      .finish
        push    eax
        stdcall StrToNumEx, eax
        stdcall StrDel ; from the stack

.finish:
        return
endp


; Writes a record with the given 32bit integer number
; Returns:
;   Nothing
proc CfgSetInt, .pConfig, .path, .int
begin
        stdcall NumToStr, [.int], ntsDec
        push    eax
        stdcall CfgSetRecord, [.pConfig], [.path], eax
        stdcall StrDel ;from the stack
        return
endp


; Search for the given parameter in the database and return its value.
;
; arguments:
;   .hConfig - handle of a string with the config database.
;   .path    - pointer or handle to the string with full path to the database element.
;
; returns:
; CF=0; eax - string handle with the value of the record. if eax==0 the value is NULL.
;       ecx - the key-value separation char: ":" means directory; "=" means record.
; CF=1; the record does not exists.
;
; CfgGetRecord.__info.codesize

proc CfgGetRecord, .pConfig, .path
begin
        pushad

        call    __CfgPrepareToSearch

  if options.Threads
    .stack_offset = 8
  else
    .stack_offset = 4
  end if

        call    __CfgSearchRecord
        jc      .exit

        movzx   eax, ah
        mov     dword [esp+.stack_offset+4*regECX], eax
        mov     dword [esp+.stack_offset+4*regEAX], edx
        test    edx, edx
        jz      .exit_ok        ; NULL value

        stdcall StrExtract, esi, 0, edx
        mov     [esp+.stack_offset+4*regEAX], eax

.exit_ok:
        clc

.exit:
        call    __CfgFinalize
        popad
        return
endp




; Set the value of the given record. If the record does not exists, creates it
; with all parent directories if needed.
; Arguments:
;   .hConfig - handle of the string with the database.
;   .path    - full path to the record.
;   .hData   - pointer or handle to the string with the value for the record.
;              if == 0, cdtNULL will be written.
; Returns:
;   Nothing
;
; CfgSetRecord.__info.codesize

proc CfgSetRecord, .pConfig, .path, .hData
begin
        pushad

        call    __CfgPrepareToSearch
        push    eax

        call    __CfgSearchRecord
        jnc     .record_found

; The record with this path is not found, so it need to be created.

        mov     edi, [esp]

; search back a previous line.
.back_loop:
        cmp     esi, edi
        je      .create_here

        dec     esi
        mov     cl, byte [esi]

        cmp     cl, ' '
        jae     .back_loop

        jecxz   .create_here

        inc     esi

.create_here:
        push    esi               ; store the insert address in the stack.

        stdcall StrNew
        mov     edi, eax
        dec     edx

        cmp     byte [esi-1], ' '
        jb      .scan_path

        stdcall StrCharCat, edi, $0a0d

.scan_path:
        inc     edx

        inc     ebx     ; the next level of indent
        jz      .next_indent
        inc     ebx
.next_indent:

        mov     esi, edx          ; the remainder of the path

.key_loop:
        lodsb
        cmp     al, ':'
        je      .record
        test    al, al
        jnz     .key_loop

; record found
.record:
        dec     esi
        sub     esi, edx

        mov     ecx, ebx
        jecxz   .indent_ok

.indent_loop:
        stdcall StrCharCat, edi, ' '
        dec     ecx
        jnz     .indent_loop

.indent_ok:                      ; add the subdirectory/record name. ECX==0 here
        xchg    cl, [edx+esi]
        stdcall StrCat, edi, edx
        mov     [edx+esi], cl

        lea     edx, [edx+esi]

        mov     eax, '='
        jecxz   .suffix_ok       ; cl contains the separator char
        mov     eax, ':'+$0a0d00
.suffix_ok:

        stdcall StrCharCat, edi, eax
        test    ecx, ecx
        jnz     .scan_path

        cmp     [.hData], 0
        je      .ins_ready

        stdcall StrCat, edi, [.hData]

.ins_ready:
        stdcall StrCharCat, edi, $0a0d
        mov     [.hData], edi
        pop     esi             ; position where to insert the content.
        xor     edx, edx        ; there is no old record.
        mov     [.path], edx    ; the [.hData] must be freed after use.

; The record with this path was found, so, the old value have to be deleted and
; new one to be recorded.

.record_found:
; here, esi == address, where the value should be inserted.
;       edx == length of the old record value, that have to be deleted.

; search the bounds of the record...
        pop     edi             ; start address of the config file.
        sub     esi, edi        ; position of the record start.
        mov     ecx, [.pConfig]

        stdcall StrSplit, [ecx+TConfig.hConfig], esi

        test    edx, edx
        jz      .record_ok      ; no need to delete the old record.

        push    eax
        stdcall StrSplit, eax, edx
        stdcall StrDel ; from the stack

.record_ok:
        mov     edx, eax        ; the part of the string after the record.

        cmp     [.hData], 0
        je      .restore

        stdcall StrCat, [ecx+TConfig.hConfig], [.hData]

.restore:
        stdcall StrCat, [ecx+TConfig.hConfig], edx

        stdcall StrDel, eax
        stdcall StrDel, edx

        cmp     [.path], 0
        jne     .exit_ok

        stdcall StrDel, [.hData]

.exit_ok:
        call    __CfgFinalize
        popad
        return
endp


; Reads the a config directory and places all record keys as strings in TArray of type dword
; Only the records of the specified directory are read. The subdirectory records ends with ":"
;
; CfgGetDir.__info.codesize

proc CfgGetDir, .hConfig, .dir
.start dd ?
begin
        pushad

        call    __CfgPrepareToSearch

  if options.Threads
    .stack_offset = 8
  else
    .stack_offset = 4
  end if

        call    __CfgSearchRecord
        jc      .finish

        sub     esi, ecx

        stdcall CreateArray, 4
        mov     edx, eax

.outer_loop:
; compute the line indent
        mov     ecx, esi

.scan_indent:
        lodsb
        test    al, al
        jz      .end_directory

        cmp     al, ' '
        je      .scan_indent
        jb      .outer_loop

        cmp     al, ';'
        jne     .indent_ok

; skips to the end of the line.
.SkipIt:
        lodsb
        test    al, al
        jz      .end_directory

        cmp     al, ' '
        jae     .SkipIt
        jmp     .outer_loop

.end_directory:
        dec      esi
        mov      [esp+.stack_offset+4*regEAX], edx
        clc

.finish:
        call    __CfgFinalize
        popad
        return

.indent_ok:
        dec     esi

        mov     [.start], esi

        sub     ecx, esi
        neg     ecx

; Here the indent of the current line is in ECX.
        cmp     ecx, ebx
        jle     .end_directory  ; if the current indent is less of equal than the block
                                ; indent, then the directory ended.
        cmp     ecx, edi
        ja      .SkipIt         ; the current indent is above the current upper limit, so
                                ; the whole line have not to be processed.

        or      edi, -1

; search the end of line and check for directory.
.dir_loop:
        lodsb
        test    al, al
        jz      .end_directory

        cmp     al, ' '
        jb      .outer_loop

        cmp     al, '='
        je      .copy_record

        cmp     al, ':'
        jne     .dir_loop

.skip_dir:
        mov     edi, ecx        ; it is a subdirectory, that is not in the path,
                                ; so set the upper limit.
        inc     esi

.copy_record:
        dec     esi
        mov     eax, esi
        sub     eax, [.start]
        jle     .SkipIt

        stdcall StrExtract, [.start], 0, eax
        push    eax
        stdcall AddArrayItems, edx, 1
        pop     dword [eax]

        jmp     .SkipIt
endp



; removes the record [.path] from the configuration database [.hConfig]
; returns CF=1 when the record does not exists.

proc CfgDelRecord, .pConfig, .path
begin
        pushad

        call    __CfgPrepareToSearch
        push    eax

        call    __CfgSearchRecord
        pop     edi
        jc      .exit

        lea     edx, [esi+edx]

.forward:
        cmp     byte [edx], 0
        je      .backward
        cmp     byte [edx], ' '
        lea     edx, [edx+1]
        jb      .forward

        dec     edx

.backward:
        cmp     esi, edi
        je      .cut_from

        dec     esi
        cmp     byte [esi], ' '
        jae     .backward

        inc     esi

.cut_from:
        sub     edx, esi
        sub     esi, edi
        mov     ecx, [.pConfig]

        stdcall StrSplit, [ecx+TConfig.hConfig], esi
        push    eax
        stdcall StrSplit, eax, edx
        stdcall StrDel ; from the stack

        stdcall StrCat, [ecx+TConfig.hConfig], eax
        stdcall StrDel, eax
        clc

.exit:
        call    __CfgFinalize
        popad
        return
endp


; Checks whether a record with path [.path] exists in the config database.
; Returns:
;   CF=0 if the record exists.
;   CF=1 if the record does not exists.
;
; CfgRecordExists.__info.codesize

proc CfgRecordExists, .hConfig, .path
begin
        pushad
        call    __CfgPrepareToSearch
        call    __CfgSearchRecord
        call    __CfgFinalize
        popad
        return
endp


; __CfgPrepareToSearch.__info.codesize

proc __CfgPrepareToSearch
begin

  virtual at ebp+8
    .pConfig dd ?
    .path    dd ?
  end virtual

        pop     ecx

        mov     esi, [.pConfig]

  if options.Threads
        lea     eax, [esi+TConfig.mutex]
        push    eax                             ; for later release
        stdcall WaitForMutex, eax, -1
  end if

        stdcall StrDup, [esi+TConfig.hCurrentDir]
        cmp     [.path], 0
        je      @f
        stdcall StrCat, eax, [.path]
@@:
        push    eax                             ; for later free

        stdcall StrPtr, eax
        mov     edx, eax

        stdcall StrPtr, [esi+TConfig.hConfig]
        mov     esi, eax

        push    ecx     ; the return address
        return
endp


proc __CfgFinalize
begin
        pop     ecx
        stdcall StrDel ; from the stack

if options.Threads
        stdcall MutexRelease ; from the stack
end if
        push    ecx
        return
endp


; __CfgSearchRecord.__info.codesize
;
; Arguments:
;   esi - pointer to the text of the database
;   edx - pointer to the path string. If the path ends with ":" the function search for the
;         first record in the given directory. and returns pointer to the start of the record key.
; Returns:
;   CF=0 if the path is found.
;     esi - pointer to the value.
;     ebx - indent of the directory. All records has to be indented more than this value.
;     ecx - indent of the record.
;     edx - length of the data record.
;     ah=":" if the path is directory.
;     ah="=" if the path is data field.
;
;   CF=1 the path was not found.
;     esi - points to the first row after the end of the block where the path
;           is supposed to be (but is not).
;     edx - points to the remainder of the path that was not found.
;     ebx - the indent of the block where the record should be placed.
proc __CfgSearchRecord
begin
        mov     edi, edx
        or      ebx, -1                 ; current block indent starts with -1

.outer_loop:
; compute the line indent
        mov     ecx, esi

.scan_indent:
        lodsb
        test    al, al
        jz      .end_of_file

        cmp     al, ' '
        je      .scan_indent
        jb      .outer_loop

        cmp     al, ';'
        jne     .indent_ok

; skips to the end of the line.
.SkipIt:
        lodsb
        cmp     al, ' '
        jae     .SkipIt
        test    al, al
        jnz     .outer_loop

.end_of_file:
;        dec      esi

.error:
        stc
        return

.indent_ok:
        dec     esi

        sub     ecx, esi
        neg     ecx

; Here the indent of the current line is in ECX.

        cmp     ecx, ebx
        jle     .error          ; if the current indent is less of equal than the block
                                ; indent, then the needed element can not be found.
        cmp     ecx, edi
        ja      .SkipIt         ; the current indent is above the current upper limit, so
                                ; the whole line have not to be processed.

        mov     edi, edx        ; the start of the current path (it is always greater than the current indent)

        cmp     byte [edi], 0
        je      .found          ; the first byte is 0, so the path was found and it is an directory, not a record.

.key_loop:
        mov     al, [edi]
        inc     edi
        mov     ah, [esi]
        inc     esi

        test    al, al
        jz      .end_of_path

        cmp     al, ':'
        je      .key_maybe_found

        test    ah, ah
        jz      .end_of_file

        cmp     al, ah
        je      .key_loop

; not this, so check the current line whether it is a directory.
.not_this:
        dec     esi

.dir_loop:
        lodsb
        test    al, al
        jz      .end_of_file

        cmp     al, ' '
        jb      .outer_loop
        je      .SkipIt
        cmp     al, '='
        je      .SkipIt
        cmp     al, ':'
        jne     .dir_loop

.skip_dir:
        mov     edi, ecx        ; it is a subdirectory, that is not in the path, so set the upper limit.
        jmp     .SkipIt

.end_of_path:
        cmp     ah, '='
        je      .found

.key_maybe_found:
        cmp     ah, ':'
        jne     .not_this

        mov     edx, edi        ; the subdirectory was found. Set new path and block indent.
        mov     ebx, ecx

        test    al, al
        jnz     .SkipIt

.found:                         ; the whole path was found, so now search for the length
        mov     edx, esi

.end_loop:
        mov     al, [edx]
        inc     edx
        cmp     al, ' '
        jae     .end_loop

        stc
        sbb     edx, esi
        return
endp





endmodule
